import java.applet.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.awt.datatransfer.*;

/**
 * <p>
 * Mod 20.03.2001 added  at the end of kochString, 
 * corrected spelling of Farnsworth in aboutMorseTraining string
 * added RandomLetters as option
 *
 * MorsePractice helps you learn Morse Code. You can use it to
 * train your ear to recognize the individual symbols, to listen
 * to randomly-generated QSO's (amateur radio contacts), or to
 * listen to any text you can type or read from a file. You can
 * also save the audio file MorsePractice generates from its
 * messages.
 * </p>
 * <p>
 * Random symbol training is based on the Koch training sequence,
 * described by David Finley in
 * <a href="http://griffy.nmt.edu/sara/sara/finley.morse.html">
 * <code>&lt;http://griffy.nmt.edu/sara/sara/finley.morse.html&gt;</code>
 * </a>
 * and
 * <a href="http://griffy.nmt.edu/sara/sara/finley.sm404.html">
 * <code>&lt;http://griffy.nmt.edu/sara/sara/finley.sm404.html&gt;</code>
 * </a>.
 * </p>
 * <p>
 * The Koch method is also described in the book, "The Art and Skill of
 * Radio-telegraphy," by William G. Peirpoint, N0HFF, as
 * distributed on the Web at
 * <a href="www.joates.demon.co.uk/megs/N0HFF/index.htm">
 * <code>&lt;www.joates.demon.co.uk/megs/N0HFF/index.htm&gt;</code>
 * </a>
 * </p>
 * <p>
 * According to the Kochmethod, you should learn the code in the
 * following sequence:
 * <blockquote>
 *	    <code>K M R S U A P T L O</code>
 *	<br><code>W I . N J E F 0 Y ,</code>
 *	<br><code>V G 5 / Q 9 Z H 3 8</code>
 *	<br><code>B ? 4 2 7 C 1 D 6 X</code>
 *	<br><code>&lt;BT&gt; &lt;SK&gt; &lt;AR&gt;</code>
 * </blockquote>
 * </p>
 * <p>
 * The three "prosigns", <code>BT</code>, <code>AR</code>,
 * and <code>SK</code> are representated by the characters '=', '+', and '#'
 * respectively.
 * </p>
 * <p>
 * The QSO format and content are based on the SuperiorMorse program
 * written by Paul J. Drongowski (N2OQT). The random sentence generator
 * is based on one that I wrote for the Decus C compiler toolkit which,
 * in turn, was based on James Gimpel's "Algorithms in SNOBOL4."
 * </p>
 * <p>
 * The Sun Audio package is described (thanks, Sherlock) in
 * <a href="http://www.cooper.edu/~donahu/auformat/auFormat.html">
 * <code>&lt;http://www.cooper.edu/~donahu/auformat/auFormat.html&gt;</code>
 * </a>.
 * </p>
 * <p>
 * Additional information on audio synthesis is available in
 * <br><a href="http://www.alulabs.com/fordevelopers/article3.html">
 *   <code>&lt;http://www.alulabs.com/fordevelopers/article3.html&gt;</code>
 * </a>,
 * <br><a href="http://www.alulabs.com/fordevelopers/PlayingWave/PlayingWave.java">
 *   <code>&lt;http://www.alulabs.com/fordevelopers/PlayingWave/PlayingWave.java&gt;</code>
 * </a>,
 * <br><a href="http://java.sun.com/products/java-media/mail-archive/Framework/0298.html">
 *   <code>&lt;http://java.sun.com/products/java-media/mail-archive/Framework/0298.html&gt;</code>
 * </a>,
 * <br>and <a href="http://www.acme.com/resources/classes/Acme/SynthAudioClip.java">
 *   <code>&lt;http://www.acme.com/resources/classes/Acme/SynthAudioClip.java&gt;</code>
 * </a>.
 * </p>
 * <p>
 * Morse code sending speed is measured in words per minute, where the
 * standard word is "PARIS". The duration of the code elements (dits and dahs)
 * are defined in "ticks" as follows (a tick is essentially equal to one
 * Baud):
 * <ul>
 *  <li>a dit is one tick followed by a one tick silence.
 *  </li>
 *  <li>a dah is three ticks followed by a one tick silence..
 *  </li>
 *  <li>a letter is a sequence of dits and dahs followed by
 *	  three ticks (total, i.e., the normal one tick with two
 *	  additional ticks).
 *  </li>
 *  <li>a word is a sequence of letters followed by 7 ticks (the
 *	  normal 3 tick letter silence and 4 additional ticks).
 * </ul>
 * Thus, the word "PARIS" "<b>.--. .- .-. .. ...</b>" contains
 * 10 dits, 4 dahs, 5 letters, and 1 word, for a total of 50 Baud.
 * To convert words per minute to milliseconds, then:
 * <br>	msec/tick = (60 * 1000) / (wpm * 50)
 * <br>&nbsp;&nbsp;5 wpm&nbsp;240 msec
 * <br>&nbsp;13 wpm&nbsp;92 msec
 * <br>&nbsp;18 wpm&nbsp;67 msec
 * <br>&nbsp;20 wpm&nbsp;60 msec
 * <br>
 * Farnsworth timing helps the student learn Morse Code by sending the
 * letters at a high speed (perhaps 18 wpm) and adding silence betewen
 * letters and words to lower the aggregate speed to the target (perhaps
 * 13 wpm). This is described further in MorseCodeSender.java.
 * </p>
 * <p>
 * Copyright &copy; 1999-2000 
 *	<a href="mailto:minow@pobox.com">Martin Minow</a>.
 *	All Rights Reserved.
 * </p>
 * <p>
 * <small>
 * Permission to use, copy, modify, and redistribute this software and its
 * documentation for personal, non-commercial use is hereby granted provided that
 * this copyright notice and appropriate documentation appears in all copies. This
 * software may not be distributed for fee or as part of commercial, "shareware,"
 * and/or not-for-profit endevors including, but not limited to, CD-ROM collections,
 * online databases, and subscription services without specific license.  The
 * author makes no expressed or implied warranty of any kind and assumes no
 * responsibility for errors or omissions. No liability is assumed for any incidental
 * or consequental damages in connection with or arising out of the use of the
 * information or program.
 * </small>
 * </p>
 *
 * @author <a href="mailto:minow@pobox.com">Martin Minow</a>
 * @version 1.1
 * @since This application requires full Java 1.1 support, but
 *	does not use Swing or other Java 1.2 capabilities. It
 *	was written and tested on a Macintosh running System
 *	8.6 and MRJ 2.1.4 supporting Java 1.1.8. Except for
 *	the capability to use Mac-specific classes to set
 *	file type and creator, and the need for the non-standard
 *	Sun Audio classes, this application passes the static
 *	100% pure Java test suite as of Aug. 21, 1999.
 */
public class MorsePractice extends Applet
	implements	ActionListener,		/* Menu and button events	*/
			ItemListener,		/* Checkbox listener		*/
			AudioSynthesizer.Listener, /* Synthesis active		*/
			EditTextAreaListener,	/* TextArea with Edit Menus	*/
			ComponentListener	/* Watch for open/close events	*/
{
    public static final boolean DEBUG                   = false;
    /*
     * These are default settings for the configuration dialogs. They are
     * superseded by the user's parameter file.
     */
    public static final int	defaultRandomDuration	= 1;	/* Minutes		*/
    public static final boolean defaultFiveCharGroups	= true;
    public static final String	defaultParamFile	= "MorsePractice.txt";
    /*
     * If we use the Farnsworth method, code sent at less than the Farnsworth
     * rate is sent is sent with the symbols at the Farnsworth rate and the
     * symbol and letter gaps enlongated to bring the overall rate down to
     * the text rate..
     */
    public static final boolean defaultUsingFarnsworth	= true;
    public static final int 	defaultTextRate 	= 13;	/* Words per minute	*/
    public static final int	defaultFarnsworthRate	= 18;	/* Words per minute	*/
    /*
     * To avoid glitches, the program always prepends a bit of silence before the
     * first symbol to let the synthesizer settle.
     */
    public static final long	initialSilence		= 50;	/* Msec			*/
    
    public static final String	author			= "Martin Minow, <minow@pobox.com>";
    /**
     * kochString defines the Koch training sequence.
     */
    public static final String kochString	=
    	"KMRSUAPTLOWI.NJEF0Y,VG5/Q9ZH38B?427C1D6X=+#";
    /**
     * kochChars is the Koch training sequence as a character vector.
     */
    public static final char kochChars[] = new char[kochString.length()];
    {
	for (int i = 0; i < kochChars.length; i++) {
	    kochChars[i]	= MorsePractice.kochString.charAt(i);
	}
    }
    public static final String	aboutMorseTraining	=
	"To learn Morse Code using random symbols and the Koch method,"
	+ " train in this order:"
	+ "\n\u00A0\u00A0 K M R S U A P T L O"
	+ "\n\u00A0\u00A0 W I . N J E F 0 Y ,"
	+ "\n\u00A0\u00A0 V G 5 / Q 9 Z H 3 8"
	+ "\n\u00A0\u00A0 B ? 4 2 7 C 1 D 6 X"
	+ "\n\u00A0\u00A0 BT or ="
	+ "\n\u00A0\u00A0 SK or +"
	+ "\n\u00A0\u00A0 AR or #"
	+ "\n\n"
	+ "Set the code speed fairly high using \"Farnsworth\" spacing"
	+ " and train on the selected characters until you are at least"
	+ " 90% correct for a 5 minute sequence, then add the next character."
	+ " The Koch Training Sequence field shows the characters in the"
	+ " recommended training order."
	+ "\n\n"
	+ "To start training, select the symbol set, sending speed, and"
	+ " training duration, then click on the \"Random Trial\" button."
	+ " You will hear five short beeps, before the training begins." 
	+ " (If you don't select any symbol, the program will use"
	+ " the entire Morse Code set.)"
	+ "\n\n"
	+ "After the training completes, type what you heard into the"
	+ " text box, using '~' as a placeholder for characters you"
	+ " missed, then click on the \"Check\" button to see the"
	+ " results."
	+ "\n\n"
	+ "The program can save your training results in a \"student"
	+ " workbook\" file. (Analysis programs haven't been written yet.)"
	+ " MorsePractice saves your training set and speed parameters in a"
	+ " parameter file named MorsePractice.txt."
	+ "\n\n"
	+ "You can also listen to QSO's (amateur radio contacts) generated"
	+ " using a random grammar."
	+ " These follow the format used in the United States for"
	+ " license examinations. When you select the \"Random QSO\""
	+ " option, MorsePractice will synthesize a QSO, then display"
	+ " it in the text area. By the way, the callsigns"
	+ " are not necessarily real and don't always match the"
	+ " license class: a novice might have an \"amateur extra\""
	+ " license. Similarly, you should expect other inconsistencies,"
	+ " such as a 13 year-old retired politician. The program uses"
	+ " your current sending speed to select between a short,"
	+ " medium, and long QSO, or you can select one of the three"
	+ " lengths directly."
	+ "\n\n"
 	+ "You will have to check your listening ability yourself: the"
	+ " \"Check\" button only works for random symbol training"
	+ " since, for high speeds, you should be listening for content,"
	+ " and will not be concerned with the exact symbols that"
	+ " MorsePractice sends."
	+ "\n\n"
	+ "Additional random symbol training options concentrate"
	+ " on symbols of various lengths, digits, and punctuation and"
	+ " prosigns, or amateur call signs. After listening to the"
	+ " sequence, type what you heard and click the \"Check\" button"
	+ " to get your score."
	;
    public static final String	aboutMorsePractice	=
    	"MorsePractice lets you listen to Morse Code or learn the code"
	+ " by listening to randomly-generated code symbols or amateur"
	+ " contacts (QSO's)."
	+ "\n\n"
	+ "You can use the large text area to listen to text you choose,"
	+ " including text you load from a file or the system clipboard."
	+ " To enter \"prosigns\" such as AR, you must surround each by"
	+ " spaces or, preferably, use the single character equivalents:"
	+ "\n\u00A0\u00A0 BT or ="
	+ "\n\u00A0\u00A0 SK or +"
	+ "\n\u00A0\u00A0 AR or #"
	+ "\n\n"
	+ "The \"File\" Menu lets you load text into the text area,"
	+ " save text to a file, or save an audio file from Morse"
	+ " Code that you generated."
	+ "\n\n"
	+ "MorsePractice provides two kinds of random message practice:"
	+ " random symbol and random QSO (\"QSO\" is an amateur"
	+ " radio abbreviation for a contact between two amateur"
	+ " radio operators). Click the \"Training Info\" button"
	+ " for more information about random symbol and random"
	+ " QSO training." 
	+ "\n\n"
	+ "Programmed by Martin Minow,  minow@pobox.com\n"
	+ "Copyright \u00A9 1999-2000, Martin Minow. All Rights Reserved."
	;
    /*
     * Configure the user Interaction
     */
    public static final int BORDER_THICKNESS	= 3;
    public static final int BORDER_GAP		= 1;
    Font dialogFont			= new Font("Dialog", Font.PLAIN, 12);
    Font symbolFont			= new Font("Monospaced", Font.PLAIN, 12);
    Font kochFont			= new Font("Monospaced", Font.BOLD, 14);
    private boolean isJustCreated	= true;
    private EditTextArea textArea	= null;
    private KochTextField kochTextField	= null;
    private static final String	pRandomSymbols	= "Random Symbols";
    private static final String pRandomQSO	= "Random QSO";
    private static final String pRandomShortQSO = "Random Short QSO";
    private static final String pRandomMediumQSO = "Random Medium QSO";
    private static final String pRandomLongQSO  = "Random Long QSO";
    private static final String pRandomDigits   = "Random Digits";
    private static final String pRandomLetters  = "Random Letters";
    private static final String pRandom2Element = "Random 1 & 2-Elem Symbs";
    private static final String pRandom3Element = "Random 3-Elem Symbs";
    private static final String pRandom4Element = "Random 4-Elem Symbs";
    private static final String pRandomPunct    = "Random Punct & Signs";
    private static final String pRandomCallSign = "Random Call Signs";
    
    private Choice randomChoice		= new Choice();
    {
    	randomChoice.addItem(pRandomSymbols);
    }
    private Button listenButton		= new Button(" Listen ");
    private Button randomButton		= new Button(" Start Trial ");
    private Button stopButton		= new Button(" Stop ");
    private Button checkButton		= new Button(" Check ");
    private Button trainButton		= new Button(" Training Info ");
    private Button infoButton		= new Button(" Program Info ");
    private IntegerField textRateField	= new IntegerField(
    		3, defaultTextRate, 0, 50);
    private IntegerField farnsworthRateField	= new IntegerField(
    		3, defaultFarnsworthRate, 0, 50);
    private IntegerField randomDurationField	= new IntegerField(
    		3, defaultRandomDuration);
    private Checkbox usingFarnsworth	= new Checkbox("Use Farnsworth Timings");
    private Checkbox fiveCharGroups	= new Checkbox("Random Uses 5 Character Groups");
    
    /*
     * The entity names (of the form "pXXX") define the values that are stored
     * in the parameter file. The random training flavor uses the actual
     * Choice strings, defined above.
     */
    public static final String pTextRate	= "TextRate";
    public static final String pFarnsworthRate	= "FarnsworthRate";
    public static final String pTrialDuration	= "TrialDuration";
    public static final String pUseFarnsworth	= "UseFarnsworth";
    public static final String pUse5CharGroups	= "UseFiveCharGroups";
    public static final String pRandomChars	= "RandomCharacters";
    public static final String pRandomFlavor	= "RandomFlavor";
    
    /*
     * Menus
     */
    private MenuBar menuBar		= new MenuBar();
    /*
     * The File Menu
     */
    private Menu fileMenu		= new Menu("File");
    private MenuItem miOpen		= new MenuItem(
    			"Open...",
    			new MenuShortcut(KeyEvent.VK_O)
    		    );
    private MenuItem miSaveText		= new MenuItem(
    			"Save Text...",
    			new MenuShortcut(KeyEvent.VK_S)
    		    );
    private MenuItem miSaveAudio	= new MenuItem(
    			"Save Audio..."
    		    );
    private MenuItem miQuit		= new MenuItem(
    			"Quit",
    			new MenuShortcut(KeyEvent.VK_Q)
    		);
    private MenuItem miHelp		= new MenuItem(
    			"MorsePractice Help"
    		);
    private MenuItem miParis		= new MenuItem(
    			"Paris text for debug"
    		);

    /*
     * The Edit Menu
     */
    private Menu editMenu		= new Menu("Edit", true); /* Tearoff ok	*/
    private MenuItem miUndo		= new MenuItem(
    			"Undo",
    			new MenuShortcut(KeyEvent.VK_Z)
    		);
    private MenuItem miCut		= new MenuItem(
    			"Cut",
    			new MenuShortcut(KeyEvent.VK_X)
    		);
    private MenuItem miCopy		= new MenuItem(
    			"Copy",
    			new MenuShortcut(KeyEvent.VK_C)
    		);
    private MenuItem miPaste		= new MenuItem(
    			"Paste",
    			new MenuShortcut(KeyEvent.VK_V)
    		);
    private MenuItem miClear		= new MenuItem(
    			"Clear"
    		);
    private MenuItem miSelectAll	= new MenuItem(
    			"Select All"
    		);
    private MenuItem miSelectNone	= new MenuItem(
    			"Select None"
    		);
    /*
     * The Edit menu needs to know which text area is being edited.
     * (Currently, there's only one text area.)
     */
    private EditTextArea activeTextArea	= null;

    /*
     * The user experience is contained in the following panels:
     *	codePanel	Buttons for each Morse Code symbol
     *	textEntryPanel	Type-in text, code speed (regular and Farnsworth),
     *			listen and stop buttons.
     *	randomTextPanel	The random text field, minutes of text,
     *			Five-character code groups, and "random" button.
     *	infoButton	The info button goes into a panel by itself.
     */    
    private Panel	codePanel;
    private Panel	buttonPanel;
    private Panel	optionPanel;
    /**
     * The Help dialogs.
     */
    private AlertDialog trainDialog	= null;
    private AlertDialog	aboutDialog	= null;

    /*
     * The MorseCode database. buttons contains the Morse Code
     * checkbox using the letter as key.
     */
    private Hashtable buttons		= new Hashtable();
    private MorseCode morseSpace	= new MorseCode(' ', " ");
    private MorseCodeSender morseSender	= new MorseCodeSender();
    
    /*
     * randomQSO contains the random sentence grammar.
     */
    private RandomQSO randomQSO		= null;
    
    /*
     * The most recent random text sequence is stored here.
     * wasRandomQSO is true to load the randomText string
     * into the text buffer when the code stops playing.
     * This is only needed for random QSO's as the user
     * types in what was received for Koch training, and
     * the text should already be present for "listen
     * to the text". This is a bit messy and could stand
     * some cleanup.
     */
    private String	randomText	= "";
    private boolean	wasRandomQSO	= false;
    
    /*
     * This is set true if MorsePractice is started as an application.
     */
    private boolean	isApplication	= false;
    
    public MorsePractice()
    {
	this(false);
    }
    
    /**
     * This is a special constructor that is called with
     * isApplication set true when MorsePractice runs as an
     * application. It's effect is to enable the File menu.
     * See the main() method below.
     * @param isApplication	true if an application, false if a real applet.
     */
    public MorsePractice(
	    boolean		isApplication
	)
    {
	super();
	this.isApplication	= isApplication;
        // MorsePractice.log("Hello world!");
    }
    /**
     * Initialize the MorsePractice applet/application.
     */
    public void init()
    {
	if (AudioSynthesizer.isSupported() == false) {
	    /*
	     * Since MorsePractice uses sun.audio, it will fail
	     * if run by  a browser that supports AudioClip using
	     * a different audio implementation.
	     */
	    new AlertDialog(
		    this,
		    "Morse Practice Failure",
		    "MorsePractice requires the \"sun.audio\" classes."
		    + "\n"
		    + "Unfortunately, these are not supported on this"
		    + " Java implementation. The applet will exit when"
		    + " you click on the \"OK\" button.",
		    true
		);
	    return;
	}
	try {
	    randomQSO           = new RandomQSO();
            randomChoice.addItem(pRandomQSO);
            randomChoice.addItem(pRandomShortQSO);
            randomChoice.addItem(pRandomMediumQSO);
            randomChoice.addItem(pRandomLongQSO);
            randomChoice.addItem(pRandomDigits);
            randomChoice.addItem(pRandomLetters);
            /* randomChoice.addItem(pRandom2Element); -- not needed? */
            randomChoice.addItem(pRandom3Element);
            randomChoice.addItem(pRandom4Element);
            randomChoice.addItem(pRandomPunct);
            randomChoice.addItem(pRandomCallSign);
	}
	catch (Exception e) {
	    System.out.println("Can't create random QSO grammar: " + e.toString());
	}
	addComponentListener(this);
	setBackground(Color.lightGray);
	createCodePanel();
	createButtonPanel();
	createOptionPanel();
	/*
	 * Create the KochTextField after creating the code panel.
	 */
	kochTextField		= new KochTextField(this);
	kochTextField.setFont(kochFont);
	kochTextField.setBackground(getBackground());
	/*
	 * Create the menus before creating the text areas because
	 * the EditTextArea may call back to enableEditMenu on
	 * initialization.
	 */
	Frame window		= AlertDialog.getFrame(this);
	textArea		= new EditTextArea(
		"", 16, 48, TextArea.SCROLLBARS_VERTICAL_ONLY, this);
	createMenus();
	fiveCharGroups.setState(defaultFiveCharGroups);
	usingFarnsworth.setState(defaultUsingFarnsworth);
	textArea.setFont(symbolFont);
	randomChoice.setFont(dialogFont);
	randomButton.setFont(dialogFont);
	randomButton.addActionListener(this);
	listenButton.setFont(dialogFont);
	listenButton.addActionListener(this);
	stopButton.setFont(dialogFont);
	stopButton.addActionListener(this);
	checkButton.setFont(dialogFont);
	checkButton.addActionListener(this);
	trainButton.setFont(dialogFont);
	trainButton.addActionListener(this);
	infoButton.setFont(dialogFont);
	infoButton.addActionListener(this);
	stopButton.setEnabled(false);
	listenButton.setEnabled(false);
	checkButton.setEnabled(false);
	morseSender.addListener(this);
	createUserExperience();
	window.setMenuBar(menuBar);
	enableEditMenu(null, false, false);	/* No edit until focus gained	*/
	if (false) {
	    Dimension d = getPreferredSize();
	    showStatus("width = " + d.width + ", height = " + d.height);
	}
	if (isApplication) {
	    readParameters();			/* After default state was set	*/
	}
    }
    /**
     * createMenus will be called by the class initializer to
     * create the various menus and add them to the MenuBar
     */
    private void createMenus()
    {
	if (isApplication) {
	    fileMenu.add(miOpen);
	    fileMenu.add(miSaveText);
	    fileMenu.addSeparator();
	    fileMenu.add(miSaveAudio);
	    fileMenu.addSeparator();
	    fileMenu.add(miQuit);
	    menuBar.add(fileMenu);
	}
	/* */
	editMenu.add(miUndo);
	editMenu.add(miCut);
	editMenu.add(miCopy);
	editMenu.add(miPaste);
	editMenu.addSeparator();
	editMenu.add(miClear);
	editMenu.addSeparator();
	editMenu.add(miSelectAll);
	editMenu.add(miSelectNone);
	if (false) {	/* Debugging */
	    editMenu.addSeparator();
	    editMenu.add(miParis);
	}
	/* */
	menuBar.add(editMenu);
	/* */
	Menu helpMenu	= menuBar.getHelpMenu();
	if (helpMenu != null) {
	    helpMenu.addSeparator();
	}
	else {
	    helpMenu	= new Menu("Help", true);
	    menuBar.setHelpMenu(helpMenu);
	}
	helpMenu.add(miHelp);
	miHelp.addActionListener(this);
	for (int i = 0; i < fileMenu.getItemCount(); i++) {
	    fileMenu.getItem(i).addActionListener(this);
	}
	for (int i = 0; i < editMenu.getItemCount(); i++) {
	    editMenu.getItem(i).addActionListener(this);
	}
    }
    /**
     * createUserExperience will be called by the class
     * initializer to organize the components on the display.
     */
    private void createUserExperience()
    {
	/*
	 * Sew the display components together.
	 * The code panel is at the upper-left corner.
	 */
	GridBagLayout gridbag	= new GridBagLayout();
	GridBagConstraints c	= new GridBagConstraints();
	setLayout(gridbag);
	c.gridy			= 0;
	c.gridx			= 0;
	c.gridwidth		= 1;
	c.gridheight		= 1;
	c.weightx		= 1.0;
	c.weighty		= 0.0;
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.NORTHWEST;
	c.insets		= new Insets(2, 2, 2, 2);
	gridbag.setConstraints(codePanel, c);
	add(codePanel);
	/*
	 * The button panel is at the upper-right corner
	 */
	c.gridx			= 1;
	c.fill			= GridBagConstraints.VERTICAL;
	c.anchor		= GridBagConstraints.NORTHEAST;
	gridbag.setConstraints(buttonPanel, c);
	add(buttonPanel);
	/*
	 * The text entry panel -- full width, fill both ways.
	 */
	c.gridy			+= 1;
	c.gridx			= 0;
	c.gridwidth		= 2;
	c.weighty		= 1.0;
	c.anchor		= GridBagConstraints.NORTHWEST;
	c.fill			= GridBagConstraints.BOTH;
	textArea.setBackground(Color.white);
	EtchedBorder textAreaBorder =
			new EtchedBorder(textArea, "Listen or Practice Text");
	textAreaBorder.setThickness(BORDER_THICKNESS);
	textAreaBorder.setGap(BORDER_GAP);
	textAreaBorder.setBackground(getBackground());
	gridbag.setConstraints(textAreaBorder, c);
	add(textAreaBorder);
	/*
	 * The Koch text panel -- full width, no fill.
	 */
	c.gridy			+= 1;
	c.gridx			= 0;
	c.weighty		= 0.0;
	c.weightx		= 1.0;
	c.fill			= GridBagConstraints.HORIZONTAL;
	EtchedBorder kochBorder	=
			new EtchedBorder(kochTextField, "Koch Training Sequence");
	kochBorder.setThickness(BORDER_THICKNESS);
	kochBorder.setGap(BORDER_GAP);
	gridbag.setConstraints(kochBorder, c);
	add(kochBorder);
	/*
	 * The option panel -- full width, fill horizontally.
	 */
	c.gridy			+= 1;
	gridbag.setConstraints(optionPanel, c);
	add(optionPanel);
    }
    /**
     * createButtonPanel creates one of the display areas.
     */
    private void createButtonPanel()
    {
	GridBagLayout gridbag	= new GridBagLayout();
	GridBagConstraints c	= new GridBagConstraints();
    	buttonPanel		= new Panel(gridbag);
    	c.gridy			= 0;
	c.gridx			= 0;
	c.gridwidth		= 1;
	c.weightx		= 0.0;
	c.weighty		= 0.0;
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.NORTHEAST;
	c.insets		= new Insets(2, 2, 2, 2);
	gridbag.setConstraints(randomChoice, c);
	buttonPanel.add(randomChoice);
	c.gridy			+= 1;
	gridbag.setConstraints(randomButton, c);
	buttonPanel.add(randomButton);
	c.gridy			+= 1;
	gridbag.setConstraints(checkButton, c);
	buttonPanel.add(checkButton);
	c.gridy			+= 1;
	gridbag.setConstraints(listenButton, c);
	buttonPanel.add(listenButton);
	c.gridy			+= 1;
	gridbag.setConstraints(stopButton, c);
	buttonPanel.add(stopButton);
	c.gridy			+= 1;
	c.weighty		= 1.0;
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.SOUTHEAST;
	gridbag.setConstraints(trainButton, c);
	buttonPanel.add(trainButton);
	c.gridy			+= 1;
	c.weighty		= 0.0;
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.SOUTHEAST;
	gridbag.setConstraints(infoButton, c);
	buttonPanel.add(infoButton);
    }
    /**
     * The code panel has one checkbox for every Morse Code symbol.
     */
    private void createCodePanel()
    {
   	Panel codeButtons	= new Panel(new GridLayout(0, 6));
    	FontMetrics fm		= getFontMetrics(symbolFont);
	int buttonWidth		= fm.stringWidth("MMM");
	/*
	 * The buttons are much wider than I expected. Is this a
	 * bug or a feature?
	 */
	for (int i = 0; i < MorseCode.morse.length; i++) {
	    MorseCode morse	= MorseCode.morse[i];
	    Checkbox button	= new Checkbox(morse.getLabel());
	    button.setFont(symbolFont);
	    Dimension d		= button.getPreferredSize();
	    d.width		= buttonWidth;
	    button.setSize(d);
	    if (morse.getCode() != null) {
		button.addItemListener(this);
		String key	= String.valueOf(morse.getLetter());
		button.setName(Integer.toString(i));
		buttons.put(key, button);
	    }
	    else {
	    	button.setEnabled(false);
	    	button.setVisible(false);
	    }
	    codeButtons.add(button);
	}
	EtchedBorder border	= new EtchedBorder(
			codeButtons, "Morse Code Symbols"
		    );
	border.setThickness(BORDER_THICKNESS);
	border.setGap(BORDER_GAP);
	codePanel		= border;
	if (false) {
	    Enumeration e       = buttons.keys();
	    while (e.hasMoreElements()) {
	        System.out.println("\"" + e.nextElement() + "\"");
	    }
	}
    }
    /**
     * The option panel has the sending speed, and a few checkboxes.
     */
    private void createOptionPanel()
    {
	GridBagLayout gridbag	= new GridBagLayout();
	GridBagConstraints c	= new GridBagConstraints();
    	optionPanel		= new Panel(gridbag);
	c.gridy			= 0;
	c.gridx			= 0;
	c.gridwidth		= 1;
	c.weightx		= 0.0;
	c.weighty		= 0.0;
	c.insets		= new Insets(0, 2, 0, 2);	/* Left, Right	*/
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.EAST;
	Label overallLabel	= new Label("Overall speed (WPM):");
	gridbag.setConstraints(overallLabel, c);
	optionPanel.add(overallLabel);
	c.gridx			+= 1;
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.WEST;
	c.insets		= new Insets(0, 0, 0, 0);
	gridbag.setConstraints(textRateField, c);
	optionPanel.add(textRateField);
	c.gridx                 += 1;
	gridbag.setConstraints(fiveCharGroups, c);
	optionPanel.add(fiveCharGroups);
	/* */
	c.gridy			+= 1;
	c.gridx			= 0;
	c.weightx		= 0.0;
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.EAST;
	Label farnsworthLabel	= new Label("Farnsworth speed (WPM):");
	gridbag.setConstraints(farnsworthLabel, c);
	optionPanel.add(farnsworthLabel);
	c.gridx			+= 1;
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.WEST;
	c.insets		= new Insets(0, 0, 0, 0);
	gridbag.setConstraints(farnsworthRateField, c);
	optionPanel.add(farnsworthRateField);
	c.gridx                 += 1;
	gridbag.setConstraints(usingFarnsworth, c);
	optionPanel.add(usingFarnsworth);
	/* */
	c.gridy			+= 1;
	c.gridx			= 0;
	c.weightx		= 0.0;
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.EAST;
	Label randomDuration	= new Label("Random Text Duration (Minutes):");
	gridbag.setConstraints(randomDuration, c);
	optionPanel.add(randomDuration);
	c.gridx			+= 1;
	c.fill			= GridBagConstraints.NONE;
	c.anchor		= GridBagConstraints.WEST;
	c.insets		= new Insets(0, 0, 0, 0);
	gridbag.setConstraints(randomDurationField, c);
	optionPanel.add(randomDurationField);
        c.gridx                 += 1;
	Label dummyLabel5	= new Label("");
	gridbag.setConstraints(dummyLabel5, c);
	optionPanel.add(dummyLabel5);
    }
    /**
     * Return the minimum panel size so other window components
     * can be constructed. The minimum and preferred sizes are
     * identical.
     * @return the minimum window size.
     */
    public Dimension getMinimumSize()
    {
	return (getPreferredSize());
    }
    /**
     * Return the preferred panel size so other window components
     * can be constructed. If you add components to the window,
     * be sure to adjust the preferred size computation.
     * @return the preferred window size.
     */
    public Dimension getPreferredSize()
    {
 	int codeButtonWidth	= codePanel.getPreferredSize().width
 				+ buttonPanel.getPreferredSize().width;
 	int codeButtonHeight	= Math.max(
 		codePanel.getPreferredSize().height,
 		buttonPanel.getPreferredSize().height
 	    );
	int width		= Math.max(
		codeButtonWidth,
		textArea.getPreferredSize().width
	    );
	width			+= 2;
	int height		=
		codeButtonHeight
		+ textArea.getPreferredSize().height
		+ optionPanel.getPreferredSize().height
		+ 16;
	return (new Dimension(width, height));
    }
    /**
     * The Java runtime event manager calls actionPerformed whenever
     * the user clicks in one of the display components.
     * @param event	Describes the user's action.
     */
    public void actionPerformed(
    	    ActionEvent		event
    	)
    {
 	/*
 	 * Note that the File Menu is installed when MorsePractice
 	 * is run as an application.
 	 */
	Object source		= event.getSource();
	if (source == listenButton) {
	    listenToText();
	}
	else if (source == randomButton) {
	    String thisFlavor	= randomChoice.getSelectedItem();
	    if (thisFlavor == pRandomSymbols) {
		setRandomText();
		checkButton.setEnabled(true);
	    }
	    else if (thisFlavor == pRandomQSO) {
                int wpm         = textRateField.getValue();
		listenToRandomQSO(wpm);
		checkButton.setEnabled(false);
	    }
	    else if (thisFlavor == pRandomShortQSO) {
		listenToRandomQSO(5);
		checkButton.setEnabled(false);
	    }
	    else if (thisFlavor == pRandomMediumQSO) {
		listenToRandomQSO(13);
		checkButton.setEnabled(false);
	    }
	    else if (thisFlavor == pRandomLongQSO) {
		listenToRandomQSO(20);
		checkButton.setEnabled(false);
	    }
	    else if (thisFlavor == pRandomDigits
	          || thisFlavor == pRandomLetters
	          || thisFlavor == pRandom2Element
	          || thisFlavor == pRandom3Element
	          || thisFlavor == pRandom4Element
	          || thisFlavor == pRandomPunct
	          || thisFlavor == pRandomCallSign) {
	        setRandomSymbols(thisFlavor);
		checkButton.setEnabled(true);
	    }
	    else {
		System.out.println("No training choice selected");
	    }
	}
	else if (source == stopButton) {
	    morseSender.stop();
	}
	else if (source == checkButton) {
	    morseSender.stop();         /* Stop sending */
	    checkRandomTrial();
	}
	else if (source == trainButton) {
	    morseSender.stop();         /* Stop sending */
	    showAboutTrainingDialog();
	}
	else if (source == infoButton) {
	    morseSender.stop();         /* Stop sending */
	    showAboutProgramDialog();
	}
	else if (source == miOpen) {
	    morseSender.stop();         /* Stop sending */
	    readTextFile();
	}
	else if (source == miSaveText) {
	    morseSender.stop();         /* Stop sending */
	    saveTextFile();
	}
	else if (source == miSaveAudio) {
	    morseSender.stop();         /* Stop sending */
	    saveAudio();
	}
	else if (source == miQuit) {
            morseSender.stop();
	    writeParameters();
	    System.exit(0);
	}
	else if (source == miUndo) {
	    if (activeTextArea != null) {
		activeTextArea.performUndo();
	    }
	}
	else if (source == miCut) {
	    if (activeTextArea != null) {
		activeTextArea.cutClipboardContents();
	    }
	}
	else if (source == miCopy) {
	    if (activeTextArea != null) {
		activeTextArea.copyClipboardContents();
	    }
	}
	else if (source == miPaste) {
	    if (activeTextArea != null) {
		activeTextArea.pasteClipboardContents();
	    }
	}
	else if (source == miClear) {
	    if (activeTextArea != null) {
		activeTextArea.clearSelection();
	    }
	}
	else if (source == miSelectAll) {
	    for (int i = 0; i < kochChars.length; i++) {
		setMorseButtonState(kochChars[i], true, false);
	    }
	}
	else if (source == miSelectNone) {
	    for (int i = 0; i < kochChars.length; i++) {
		setMorseButtonState(kochChars[i], false, false);
	    }
	}
	else if (source == miParis) {
	    morseSender.testDurationComputation();
	}
	else if (source == miHelp) {
	    showAboutProgramDialog();
	}
	else {
	    System.out.println("Unknown actionEvent: " + event.toString());
	}
    }
    /**
     * The user clicked on the "Training Info" button. Display more
     * information about Koch and random QSO training.
     */
    private void showAboutTrainingDialog()
    {
	if (trainDialog == null) {
	    trainDialog = new AlertDialog(
		    this,
	    	    "Learning Morse Code",
	    	    aboutMorseTraining,
	    	    false
	    	);
	}
	trainDialog.setVisible(true);
	trainDialog.toFront();
    }
    /**
     * The user clicked on the "Info" button or the help
     * menu item. Display information about MorsePractice.
     */
    private void showAboutProgramDialog()
    {
	if (aboutDialog == null) {
	    aboutDialog = new AlertDialog(
		    this,
	    	    "About the MorsePractice Program",
	    	    aboutMorsePractice,
	    	    false
	    	);
	}
	aboutDialog.setVisible(true);
	aboutDialog.toFront();
    }
    /**
     * Implement the ItemListener Interface.
     * The user clicked on a Morse Code symbol checkbox.
     * If the user is selecting this button, make its sound
     * and add it to the random symbol training set.
     * @param event	The button that the user clicked.
     */
    public void itemStateChanged(
	    ItemEvent		event
	)
    {
	try {
	    Checkbox button	= (Checkbox) event.getItemSelectable();
	    boolean selected	= button.getState();
	    if (selected) {
                MorseCode symbol = getSymbolFromCheckbox(button);
	        listenToOneSymbol(symbol);
	    }
	}
	catch (ClassCastException e) {
	    System.out.println("Not a checkbox: " + event.toString());
	}
	catch (NumberFormatException e) {
	    System.out.println("Bogus checkbox: " + event.toString());
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }
    
    /**
     * The parameter reader calls setMorseButtonState to restore
     * the user's previous selection. This method is also called
     * by the Koch text field when the user clicks on a symbol.
     * @param letter	A Morse Code sumbol.
     * @param newState	True to select this symbol
     * @param listen    True to listen to this symbol if it's selected.
     *                  (This is set false by the parameter readers
     *                  and true by the checkbox and Koch field listeners).
     * @see #readParameters
     * @see KochTextField#mouseClicked
     */
    public void setMorseButtonState(
    	    char		letter,
    	    boolean		newState,
    	    boolean             listen
    	)
    {
	Checkbox button	= getMorseButton(letter);
	if (button != null) {
	    boolean oldState = getMorseButtonState(letter);
	    if (oldState != newState) {
		button.setState(newState);
		kochTextField.setKochLabelState(button);
		if (newState && listen) {
		    MorseCode symbol    = getSymbolFromCheckbox(button);
		    listenToOneSymbol(symbol);
		}
	    }
	}
    }
    
    /**
     * Given a Checkbox, get the MorseCodeSymbol it describes.
     * @param button            One of our MorseCode checkboxes.
     * @return The symbol this checkbox defines.
     */
    private MorseCode getSymbolFromCheckbox(
            Checkbox            button
        )
    {
	String name		= button.getName();
	int i		        = Integer.parseInt(name);
	MorseCode symbol	= MorseCode.morse[i];
	return (symbol);
    }

    /**
     * Return the selection state of a particular symbol..
     * @param letter	A Morse Code sumbol.
     * @return true if the symbol is selected.
     */
    public boolean getMorseButtonState(
	    char		letter
	)
    {
	boolean result	= false;
	Checkbox button	= getMorseButton(letter);
	if (button != null) {
	    result = button.getState();
	}
	return (result);
    }
    /**
     * Return the button that controls this symbol's
     * selection state. Return null if this is not a
     * valid symbol. The buttons are stored in a
     * Hashtable for speed.
     * @param letter	A Morse Code sumbol.
     * @return this letter's checkbox.
     */
    public Checkbox getMorseButton(
    	    char		letter
    	)
    {
	String key	= String.valueOf(letter);
	Checkbox button	= (Checkbox) buttons.get(key);
	if (button == null) {
	    System.out.println("No button for '" + letter + "', key = \"" + key + "\"");
	}
	return (button);
    }
    /**
     * Return a string consisting of the selected random
     * text characters.
     * @result The selected characters.
     */
    public String getSelectedText()
    {
	StringBuffer work	= new StringBuffer(kochChars.length);
	for (int i = 0; i < kochChars.length; i++) {
	    if (getMorseButtonState(kochChars[i])) {
		work.append(kochChars[i]);
	    }
	}
	return (work.toString());
    }
    /**
     * The user has just completed a trial and clicked the "check"
     * button. Grab what the user typed into the text area and
     * throw it at the trial display dialog.
     * @see TrialDialog
     */
    private void checkRandomTrial()
    {
	Trial trial			= new  Trial();
	long startTime                  = System.currentTimeMillis();
        // MorsePractice.log("Trial: \"" + randomText + "\"");
	trial.setTrialText(randomText);	/* What we just sent	*/
	String trialText		= textArea.getText();
	int goodCount			= 0;
	int sentCount			= 0;
	int i				= 0;
	for (; i < randomText.length(); i++) {
	    char sentChar		= Character.toUpperCase(randomText.charAt(i));
	    if (i < trialText.length()) {
		char gotChar		= Character.toUpperCase(trialText.charAt(i));
		if (fiveCharGroups.getState()
		 && sentChar == ' '
		 && gotChar == ' ') {
		    ; /* Skip the expected space */
		}
		else {
		    ++sentCount;
		    trial.addPair(sentChar, gotChar);
		    if (sentChar == gotChar) {
			++goodCount;
		    }
		}
	    }
	    else {
		++sentCount;
	    	trial.addPair(randomText.charAt(i), '*');
	    }
	}
	for (; i < trialText.length(); i++) {
	    trial.addPair('*', trialText.charAt(i));
	}
	int percent			= (goodCount * 100) / sentCount;
	String titleText                = Integer.toString(percent) + "% Correct";
	String headingText              = titleText
                        + ", "
                        + Integer.toString(goodCount)
                        + " correct, "
                        + Integer.toString(sentCount - goodCount)
                        + " incorrect, "
                        + Integer.toString(sentCount)
                        + " total";
        // MorsePractice.log("Text pairs added, " + percent + "% Correct"); 
	new TrialDialog(
		    this,
	    	    titleText,
	    	    headingText,
	    	    trial,
	    	    randomText,
	    	    trialText
	    	);
    }
    /**
     * The user clicked the Random QSO button. Synthesize a QSO
     * then load it into the text area so the user can verify
     * the results.
     */
    private void listenToRandomQSO(
            int         wpm
        )
    {
	textArea.setText("");
	randomText			= randomQSO.getQSO(wpm);
	wasRandomQSO			= true;
	listenToText(randomText, true);
    }
    /**
     * The user has just typed some text into the text area
     * and wants to listen to it.
     */
    private void listenToText()
    {
	wasRandomQSO			= false;
	listenToText(textArea.getText(), false);
    }
    /**
     * Listen to a string of text. Optionally, prefix the message
     * with five "pips" to give the user time to pick up a pencil.
     * @param text		What to say.
     * @param prefixPips	True to synthesize ticks.
     */
    private void listenToText(
	    String			text,
	    boolean			prefixPips
	)
    {
	try {
	    startSending(prefixPips);
	    morseSender.synthesize(text);
	    morseSender.silence(initialSilence);
	    // System.out.println("Playing " + sequence.size() + " symbols");
	    morseSender.play();
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }
    /**
     * Listen to an example of a Morse symbol. This is called by the
     * ItenStateListener when the user enables a Checkbox, and by
     * the setMorseButtonState when the user clicks on a Koch symbol.
     */
    protected void listenToOneSymbol(
            MorseCode           symbol
        )
    {
        try {
            if (morseSender.isSending()) {
                long sleep      = (morseSender.getDuration() + 100) / 10;
                for (int i = 0; morseSender.isSending() && i < 12; i++) {
                    Thread.sleep(sleep);
                }
            }
        }
	catch (InterruptedException e) {}
	catch (Exception e) {}
        startSending(false);
        morseSender.synthesize(symbol);
        morseSender.play();
    }

    /**
     * Build a random text sequence using the characters that the
     * user selected. This could probably be improved. In particular,
     * it would be useful to get more equal distribution, with
     * emphasis on character pairs that the student finds confusing.
     */
    private void setRandomText()
    {
	textArea.setText("");
	String kochText			= getSelectedText();
	boolean useKochText		= kochText.length() > 0;
	boolean useFiveCharGroups	= fiveCharGroups.getState();
	try {
	    startSending(true);
	    MorseCode[] selection	= MorseCode.getCodeTokens(kochText);
	    if (useKochText == false) {
		selection		= MorseCode.getCodeTokens(kochString);
	    }
	    StringBuffer work		= new StringBuffer();
	    long randomDuration		= (long) (randomDurationField.getValue()) * 60000;
	    long duration		= 0;
	    /*
	     * If the user selected five character groups, we will
	     * complete the last group. Count only counts "real"
	     * characters for 5 character groups.
	     */
	    int count			= 0;
	    for (;;) {
		MorseCode codeElement	= null;
		do {
		    int index		= (int) (Math.random() * selection.length);
		    codeElement		= selection[index];
		    /*
		     * Skip over dummy place-holders.
		     */
		    if (codeElement.getCode() == null) {
			continue;
		    }
		} while (useFiveCharGroups == false && codeElement.getLetter() == ' ');
		duration		+= morseSender.synthesize(codeElement);
		char letter		= codeElement.getLetter();
		work.append(letter);
		++count;
		if (useFiveCharGroups == false) {
		    if (letter != ' ' && duration >= randomDuration) {
			break;
		    }
		}
		else if ((count % 5) == 0) {
		    duration		+= morseSender.synthesize(morseSpace);
		    if (duration >= randomDuration) {
			break;
		    }
		    /*
		     * Appending the space after checking for duration
		     * prevents the program from checking for a trailing
		     * space, which the user won't type..
		     */
		    work.append(" ");
		}
	    }
 	    randomText			= work.toString();
	    if (false) {
		System.out.println("\"" + randomText
			+ "\" for " + duration + " msec."
		    );
	    }
	    morseSender.play();
	}
	catch (Exception e) {
	    e.printStackTrace();
	}
    }
    /**
     * Build a random text sequence using the grammar root symbol
     * indicated by the menu option.
     */
    private void setRandomSymbols(
           String               menuOption
       )
   {
        textArea.setText("");
        String root             = null;
        if (menuOption == pRandomDigits) {
            root                = "<RandomDigits>";
        }
	else if (menuOption == pRandomLetters) {
            root                = "<RandomLetters>";
        }
        else if (menuOption == pRandom2Element) {
            root                = "<Random2Element>";
        }
        else if (menuOption == pRandom3Element) {
            root                = "<Random3Element>";
        }
        else if (menuOption == pRandom4Element) {
            root                = "<Random4Element>";
        }
        else if (menuOption == pRandomPunct) {
            root                = "<RandomPunct>";
        }
        else if (menuOption == pRandomCallSign) {
            root                = "<AnyCallSign>";
        }
        else {
            System.out.println("Illegal menu option \"" + menuOption + "\"");
            root                = "<RandomDigits>";
        }
	long randomDuration     = (long) (randomDurationField.getValue()) * 60000;
	long duration		= 0;
	StringBuffer work       = new StringBuffer();
        try {
            startSending(true);
            while (duration < randomDuration) {
                String word     = randomQSO.expand(root);
                if (duration > 0) {
                    work.append(" ");
                    duration    += morseSender.synthesize(" ");
                }
                work.append(word);
                duration        += morseSender.synthesize(word);
            }
            randomText = work.toString();
            // textArea.setText(randomText);
            morseSender.play();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
            
        
    /**
     * Initialize the Morse code audio generator for a new message.
     * @param prefixPips	True to prefix the message with
     *				five short beeps.
     */
    private void startSending(
	    boolean		prefixPips
	)
    {
	morseSender.stop();		/* Paranoia			*/
	morseSender.reset();		/* Start at the beginning	*/
	morseSender.setTextRate(textRateField.getValue());
	morseSender.setFarnsworthRate(farnsworthRateField.getValue());
	morseSender.setUsingFarnsworth(usingFarnsworth.getState());
	/*
	 * Begin by sending a brief silence to let the synthesizer
	 * settle down. This helps prevent an initial click.
	 */
	try {
	    morseSender.silence(initialSilence); /* Synthesizer settle	*/
	    if (prefixPips) {
		for (int i = 0; i < 5; i++) {
		    morseSender.tick();
		}
	    }
	}
	catch (IOException e) {}
    }
    /**
     * Read a text file into the textArea.
     */
    private void readTextFile()
    {
	String text	= MorsePracticeFile.readTextFile(
				AlertDialog.getFrame(this),
				"Text to read"
			    );
	if (text != null) {
	    textArea.setText(text);
	    textArea.killUndo();
	}
    }
    /**
     * Save the current text area into a file.
     */
    private void saveTextFile()
    {
	MorsePracticeFile.saveTextFile(
		AlertDialog.getFrame(this),
		"Text to write",
		textArea.getText()
	    );
    }
    /**
     * Save the Morse Code audio into a file.
     */
    private void saveAudio()
    {
	DataOutputStream out	= MorsePracticeFile.createBinaryFile(
					AlertDialog.getFrame(this),
					"Write Audio File"
				    );
	if (out != null) {
	    try {
		morseSender.writeAudio(out);
		out.flush();
		out.close();
	    }
	    catch (IOException e) {}
	    finally {
		try {
		    if (out != null) {
			out.close();
		    }
		}
		catch (IOException e) {}
	    }
	}
    }
    
    /*
     * enableEditMenu implements the EditTextAreaListener interface.
     * It sets the Edit Menu options to conform to the text area's
     * current state.     
     * @param activeTextArea		The current text area. It is null
     *					if there is no currently-active text
     *					area.
     * @param hasSelection		True if some text is selected.
     * @param hasClipboardContent	True if there is text on the clipboard.
     */
    public void enableEditMenu(
	    EditTextArea		activeTextArea,		/* May be null	*/
	    boolean			hasSelection,
	    boolean			hasClipboardContent
	)
    {
	this.activeTextArea	= activeTextArea;
	miCut.setEnabled(hasSelection);
	miCopy.setEnabled(hasSelection);
	miClear.setEnabled(hasSelection);
	miPaste.setEnabled(hasClipboardContent);
	EditTextArea.setUndoMenuItem(activeTextArea, miUndo);
    }
    /**
     * The user just changed the active text area. Enable the listen
     * buttion if there is something to listen to.
     * @param event		The current text event. This will
     *				contain the current text area.
     */
    public void editTextValueChanged(
	   TextEvent		event
	)
    {
	listenButton.setEnabled(textArea.getText().length() != 0);
    }
    /**
     * audioSynthesis implements the AudioSynthesizerListener interface.
     * @param active	True if the program is synthesizing sound.
     */
    public void audioSynthesis(
	    boolean		active
	)
    {
	stopButton.setEnabled(active);
	if (active == false && wasRandomQSO) {
	    /*
	     * The program just stopped synthesizing a random QSO.
	     * Load the QSO into the text area so the student can
	     * check comprehension.
	     */
	    textArea.setText(randomText);
	    wasRandomQSO		= false;
	}
    }
    
    /**
     * Read the parameter file. This could be changed to use an
     * XML parser. Note that this may not be called from an Applet
     * context.
     */
    private void readParameters()
    {
	Reader reader			= null;
	LineNumberReader in		= null;
	try {
	    reader			= new FileReader(defaultParamFile);
	    in				= new LineNumberReader(reader);
	    String s;
	    while ((s = in.readLine()) != null) {
		if (s.startsWith("#") == false) {
		    /*
		     * Note: don't use = as a token separator: it's in
		     * the Morse alphabet.
		     */
		    StringTokenizer st	= new StringTokenizer(s, ":");
		    try {
			String name	= st.nextToken().trim();
			String value	= "";
			if (st.hasMoreTokens()) {
			    value	= st.nextToken().trim();
			}
			if (name.equalsIgnoreCase(pTextRate)) {
			    int rate	= Integer.parseInt(value);
			    textRateField.setValue(rate);
			}
			else if (name.equalsIgnoreCase(pFarnsworthRate)) {
			    int rate	= Integer.parseInt(value);
			    farnsworthRateField.setValue(rate);
			}
			else if (name.equalsIgnoreCase(pTrialDuration)) {
			    int rate	= Integer.parseInt(value);
			    randomDurationField.setValue(rate);
			}
			else if (name.equalsIgnoreCase(pUseFarnsworth)) {
			    boolean val	= Boolean.valueOf(value).booleanValue();
			    usingFarnsworth.setState(val);
			}
			else if (name.equalsIgnoreCase(pUse5CharGroups)) {
			    boolean val	= Boolean.valueOf(value).booleanValue();
			    fiveCharGroups.setState(val);
			}
			else if (name.equalsIgnoreCase(pRandomChars)) {
			    char[] chars	= value.toCharArray();
			    for (int i = 0; i < chars.length; i++) {
				if (chars[i] != '"') {
				    setMorseButtonState(chars[i], true, false);
				}
			    }
			}
			else if (name.equalsIgnoreCase(pRandomFlavor)) {
			    randomChoice.select(value);
			}
			else {
			    System.err.print("Unknown param \"" + s + "\" ignored");
			    System.err.println(" at line " + in.getLineNumber());
			}
		    }
		    catch (NumberFormatException e) {
			System.err.print("Number format error, \"" + s + "\" ignored");
			System.err.println(" at line " + in.getLineNumber());
		    }
		    catch (NoSuchElementException e) {
			System.err.print("Param line error, \"" + s + "\" ignored");
			System.err.println(" at line " + in.getLineNumber());
		    }
		    catch (Exception e) {
			System.err.print("Strange Param line error, \"" + s + "\" ignored");
			System.err.println(" at line " + in.getLineNumber());
		    }
		}
	    }
	}
	catch (FileNotFoundException e) { /* Ignored: no param file */ }
	catch (IOException e) {
	    System.err.println("Param file error: " + e.toString());
	}
	finally {
	    try {
		if (in != null) {
		    in.close();
		}
	    }
	    catch (Exception e) { }
	    try {
		if (reader != null) {
		    reader.close();
		}
	    }
	    catch (Exception e) {}
	}
    }
    /**
     * Write the parameter file. This could be changed to generate
     * and XML document. Note that this may not be called from
     * an Applet context
     */
    private void writeParameters()
    {
	File			outFile		= null;
	FileOutputStream	outStream	= null;
	PrintWriter		out		= null;
	try {
	    outFile		= new File(defaultParamFile);
	    outStream		= new FileOutputStream(outFile);
	    out			= new PrintWriter(outStream);
	    MacFileUtilities.setCreatorAndFileType(outFile, "R*ch", "TEXT");
	    out.println("# " + (new Date()).toString());
	    out.println(pTextRate	+ ":" + textRateField.getValue());
	    out.println(pFarnsworthRate	+ ":" + farnsworthRateField.getValue());
	    out.println(pTrialDuration	+ ":" + randomDurationField.getValue());
	    out.println(pUseFarnsworth	+ ":" + usingFarnsworth.getState());
	    out.println(pUse5CharGroups	+ ":" + fiveCharGroups.getState());
	    out.println(pRandomFlavor   + ":" + randomChoice.getSelectedItem());
	    out.print(pRandomChars + ":\"");
	    for (int i = 0; i < MorseCode.morse.length; i++) {
		char letter		= MorseCode.morse[i].getLetter();
		if (letter != '\u0000' && getMorseButtonState(letter)) {
		    out.print(letter);
		}
	    }
	    out.println("\"");
	    out.flush();
	}
	catch (IOException e) {
	    System.err.println("Error writing param file: " + e.toString());
	}
	finally {
	    try {
		if (out != null) {
		    out.close();
		    if (out.checkError()) {
			System.err.println("Error writing param file");
		    }
		}
	    }
	    catch (Exception e) {}
	    try {
		if (out != null) {
		    out.close();
		    if (out.checkError()) {
			System.err.println("Error writing param file");
		    }
		}
	    }
	    catch (Exception e) {}
	    try {
		if (outStream != null) {
		    outStream.close();
		}
	    }
	    catch (Exception e) {
		System.err.println("Error closing param file (OutputStream): " + e);
	    }
	}
	// try { Thread.sleep(5000); } catch (InterruptedException e) {}
    }
    
    /*
     * Implement the ComponentListener interface. This is a hack to
     * convince the Java environment to choose the appropriate text
     * field. requestFocus is ignored when the Component is first created.
     * Since Components that become visible when their ancestors are made
     * visible (who thought up that piece of cleverness?), we catch the
     * resized event. Currently, there is only one text area, so this
     * is overkill, but that was not true previously, and may be changed
     * in the future.
     */
    
    /**
     * componentHidden is part of the ComponentListener interface.
     * @param event	The component event.
     * @see java.awt.event.ComponentEvent
     */
    public void componentHidden(ComponentEvent event)	{ }
    /**
     * componentMoved is part of the ComponentListener interface.
     * @param event	The component event.
     * @see java.awt.event.ComponentEvent
     */
    public void componentMoved(ComponentEvent event)	{ }
    /**
     * componentShown is part of the ComponentListener interface.
     * @param event	The component event.
     * @see java.awt.event.ComponentEvent
     */
    public void componentShown(ComponentEvent event)	{ }
    /**
     * componentResized is part of the ComponentListener interface.
     * When called, it ensures that the textArea has the focus.
     * @param event	The component event.
     * @see java.awt.event.ComponentEvent
     */
    public void componentResized(
	    ComponentEvent	event
	)
    {
	if (isJustCreated) {
	    isJustCreated	= false;
	    textArea.requestFocus();
	}
    }
    /**
     * main constructs and launches the MorsePractice panel.
     * @param args command line arguments (unused)
     */
    public static void main(
	    String[]		args
	)
    {
    	try {
	    if (AudioSynthesizer.isSupported() == false) {
		throw new RuntimeException("No Audio");
	    }
	}
	catch (Exception e) {
	    new AlertDialog(
		    new Frame("Morse Code Practice"),
		    "Morse Practice Failure",
		    "MorsePractice requires the \"sun.audio\" classes."
		    + "\n"
		    + "Unfortunately, these are not supported on this"
		    + " Java implementation. The program will exit when"
		    + " you click on the \"OK\" button.",
		    true
		);
	}
	try {
	    /*
	     * On the Mac, the menus are in the menu bar, so
	     * the main window should be a bit smaller. The
	     * idea is to fit comfortably on a 640x480 display.
	     */
	    int frameHeight             = 480;
	    try {
	        String osName           = System.getProperty("os.name");
	        if (osName.startsWith("Mac")) {
	            frameHeight         -= 24;  /* Eyeball value        */
	        }
	    }
	    catch (Exception e) {}
	    Applet morsePractice	= new MorsePractice(true);
	    org.minow.applets.AppletFramework frame =
		    new org.minow.applets.AppletFramework(
			    morsePractice,
			    "MorsePractice",
			    520,
			    frameHeight,
			    args,
			    false   /* No status bar */
		     );
	}
	catch (Exception e) {
	    new AlertDialog(
		    new Frame("Morse Code Practice"),
		    "Morse Practice Failure",
		    "MorsePractice failed during execution:"
		    + "\n" + e.toString(),
		    true
		);
	}
    }
    
    /**
     * Serious debugging only
     */
    public static void log(
            String              text
        )
    {
        System.out.println(text);
        try {
            Thread.sleep(1000);
        }
        catch (InterruptedException e) {}
    }
}

